import copy
import unittest
from dataclasses import dataclass

from gpiowrapper import PinType, GPIOPinMode
from gpiowrapper.base import _Pin, _GPIOPin, PinAddressing
from gpiowrapper.raspi import Raspi40PinBarEmulator


@dataclass(frozen=True)
class RaspiPinBarTestDefaults:
    @staticmethod
    def default_pins():
        value = [
            _Pin(idx=1, type=PinType.POWER),
            _Pin(idx=2, type=PinType.POWER),
            _GPIOPin(idx=3, type=PinType.GPIO, gpio_idx=2, mode=GPIOPinMode.OFF),
            _Pin(idx=4, type=PinType.POWER),
            _GPIOPin(idx=5, type=PinType.GPIO, gpio_idx=3, mode=GPIOPinMode.OFF),
            _Pin(idx=6, type=PinType.GROUND),
            _GPIOPin(idx=7, type=PinType.GPIO, gpio_idx=4, mode=GPIOPinMode.OFF),
            _GPIOPin(idx=8, type=PinType.GPIO, gpio_idx=14, mode=GPIOPinMode.OFF),
            _Pin(idx=9, type=PinType.GROUND),
            _GPIOPin(idx=10, type=PinType.GPIO, gpio_idx=15, mode=GPIOPinMode.OFF),
            _GPIOPin(idx=11, type=PinType.GPIO, gpio_idx=17, mode=GPIOPinMode.OFF),
            _GPIOPin(idx=12, type=PinType.GPIO, gpio_idx=18, mode=GPIOPinMode.OFF),
            _GPIOPin(idx=13, type=PinType.GPIO, gpio_idx=27, mode=GPIOPinMode.OFF),
            _Pin(idx=14, type=PinType.GROUND),
            _GPIOPin(idx=15, type=PinType.GPIO, gpio_idx=22, mode=GPIOPinMode.OFF),
            _GPIOPin(idx=16, type=PinType.GPIO, gpio_idx=23, mode=GPIOPinMode.OFF),
            _Pin(idx=17, type=PinType.POWER),
            _GPIOPin(idx=18, type=PinType.GPIO, gpio_idx=24, mode=GPIOPinMode.OFF),
            _GPIOPin(idx=19, type=PinType.GPIO, gpio_idx=10, mode=GPIOPinMode.OFF),
            _Pin(idx=20, type=PinType.GROUND),
            _GPIOPin(idx=21, type=PinType.GPIO, gpio_idx=9, mode=GPIOPinMode.OFF),
            _GPIOPin(idx=22, type=PinType.GPIO, gpio_idx=25, mode=GPIOPinMode.OFF),
            _GPIOPin(idx=23, type=PinType.GPIO, gpio_idx=11, mode=GPIOPinMode.OFF),
            _GPIOPin(idx=24, type=PinType.GPIO, gpio_idx=8, mode=GPIOPinMode.OFF),
            _Pin(idx=25, type=PinType.GROUND),
            _GPIOPin(idx=26, type=PinType.GPIO, gpio_idx=7, mode=GPIOPinMode.OFF),
            _GPIOPin(idx=27, type=PinType.GPIO, gpio_idx=0, mode=GPIOPinMode.OFF),
            _GPIOPin(idx=28, type=PinType.GPIO, gpio_idx=1, mode=GPIOPinMode.OFF),
            _GPIOPin(idx=29, type=PinType.GPIO, gpio_idx=5, mode=GPIOPinMode.OFF),
            _Pin(idx=30, type=PinType.GROUND),
            _GPIOPin(idx=31, type=PinType.GPIO, gpio_idx=6, mode=GPIOPinMode.OFF),
            _GPIOPin(idx=32, type=PinType.GPIO, gpio_idx=12, mode=GPIOPinMode.OFF),
            _GPIOPin(idx=33, type=PinType.GPIO, gpio_idx=13, mode=GPIOPinMode.OFF),
            _Pin(idx=34, type=PinType.GROUND),
            _GPIOPin(idx=35, type=PinType.GPIO, gpio_idx=19, mode=GPIOPinMode.OFF),
            _GPIOPin(idx=36, type=PinType.GPIO, gpio_idx=16, mode=GPIOPinMode.OFF),
            _GPIOPin(idx=37, type=PinType.GPIO, gpio_idx=26, mode=GPIOPinMode.OFF),
            _GPIOPin(idx=38, type=PinType.GPIO, gpio_idx=20, mode=GPIOPinMode.OFF),
            _Pin(idx=39, type=PinType.GROUND),
            _GPIOPin(idx=40, type=PinType.GPIO, gpio_idx=21, mode=GPIOPinMode.OFF),
        ]
        return copy.deepcopy(value)

    @staticmethod
    def default_gpio_pins(mode: GPIOPinMode = GPIOPinMode.OFF):
        value = [
            _GPIOPin(idx=27, type=PinType.GPIO, gpio_idx=0, mode=mode),
            _GPIOPin(idx=28, type=PinType.GPIO, gpio_idx=1, mode=mode),
            _GPIOPin(idx=3, type=PinType.GPIO, gpio_idx=2, mode=mode),
            _GPIOPin(idx=5, type=PinType.GPIO, gpio_idx=3, mode=mode),
            _GPIOPin(idx=7, type=PinType.GPIO, gpio_idx=4, mode=mode),
            _GPIOPin(idx=29, type=PinType.GPIO, gpio_idx=5, mode=mode),
            _GPIOPin(idx=31, type=PinType.GPIO, gpio_idx=6, mode=mode),
            _GPIOPin(idx=26, type=PinType.GPIO, gpio_idx=7, mode=mode),
            _GPIOPin(idx=24, type=PinType.GPIO, gpio_idx=8, mode=mode),
            _GPIOPin(idx=21, type=PinType.GPIO, gpio_idx=9, mode=mode),
            _GPIOPin(idx=19, type=PinType.GPIO, gpio_idx=10, mode=mode),
            _GPIOPin(idx=23, type=PinType.GPIO, gpio_idx=11, mode=mode),
            _GPIOPin(idx=32, type=PinType.GPIO, gpio_idx=12, mode=mode),
            _GPIOPin(idx=33, type=PinType.GPIO, gpio_idx=13, mode=mode),
            _GPIOPin(idx=8, type=PinType.GPIO, gpio_idx=14, mode=mode),
            _GPIOPin(idx=10, type=PinType.GPIO, gpio_idx=15, mode=mode),
            _GPIOPin(idx=36, type=PinType.GPIO, gpio_idx=16, mode=mode),
            _GPIOPin(idx=11, type=PinType.GPIO, gpio_idx=17, mode=mode),
            _GPIOPin(idx=12, type=PinType.GPIO, gpio_idx=18, mode=mode),
            _GPIOPin(idx=35, type=PinType.GPIO, gpio_idx=19, mode=mode),
            _GPIOPin(idx=38, type=PinType.GPIO, gpio_idx=20, mode=mode),
            _GPIOPin(idx=40, type=PinType.GPIO, gpio_idx=21, mode=mode),
            _GPIOPin(idx=15, type=PinType.GPIO, gpio_idx=22, mode=mode),
            _GPIOPin(idx=16, type=PinType.GPIO, gpio_idx=23, mode=mode),
            _GPIOPin(idx=18, type=PinType.GPIO, gpio_idx=24, mode=mode),
            _GPIOPin(idx=22, type=PinType.GPIO, gpio_idx=25, mode=mode),
            _GPIOPin(idx=37, type=PinType.GPIO, gpio_idx=26, mode=mode),
            _GPIOPin(idx=13, type=PinType.GPIO, gpio_idx=27, mode=mode),
        ]
        return copy.deepcopy(value)

    @staticmethod
    def gpio_pin_ids():
        value = [
                27, 28, 3, 5, 7, 29, 31, 26, 24, 21, 19, 23, 32, 33, 8,
                10, 36, 11, 12, 35, 38, 40, 15, 16, 18, 22, 37, 13
        ]
        return copy.deepcopy(value)


class Raspi40PinBarEmulatorTests(unittest.TestCase):
    class_path = f'{Raspi40PinBarEmulator.__module__}.Raspi40PinBarEmulator'

    def test_init__Always__HasCorrectPins(self):
        # Arrange
        expected_pins = RaspiPinBarTestDefaults.default_pins()

        # Act
        bar = Raspi40PinBarEmulator(initial_addressing=PinAddressing.PinBar)

        # Assert
        self.assertListEqual(bar._pins, expected_pins)

    def test_init__Always__HasCorrectGPIOPins(self):
        # Arrange
        expected_gpio_pins = RaspiPinBarTestDefaults.default_gpio_pins()

        # Act
        bar = Raspi40PinBarEmulator(initial_addressing=PinAddressing.PinBar)

        # Assert
        self.assertListEqual(bar._gpio_pins, expected_gpio_pins)



import sys
from functools import wraps
from inspect import isclass, isfunction


class IndexableProperty:
    """
    An instance property which can be accessed and modified by pythons __getitem__ and
    __setitem__ functions. Meant to be used as decorator function.
    Implements similar interface as pythons property inbuilt function except that getter function is
    called itemgetter and setter function is called itemsetter.
    See pythons property inbuilt function for more information.
    """

    def __init__(
            self,
            fget: callable = None,
            fset: callable = None,
            fdel: callable = None,
            pdel: callable = None,
            doc: str = None
    ):
        """
        Returns an indexable property attribute.
        The advantage compared to property is, that an indexable property does not provide direct access
        to an underlying attribute and therefore favours the enclosure principle to produce cleaner code.
        E.g. if the underlying data structure is a list, you do not have access to the list object but
        instead access to the values of the list.

        :param fget: is a function for accessing attribute value(s). It's parameters have to be as __getitem__
        function.
        :param fset: is a function for setting attribute value(s). It's parameters have to be as __setitem__
        function.
        :param pdel: is a function for deleting attribute value(s). It's parameters have to be as __delitem__
        function.
        :param fdel: is a function for deleting the attribute.
        :param doc: creates a docstring for the attribute.
        """
        self.fget: callable = fget
        self.fset: callable = fset
        self.fdel: callable = fdel
        self.pdel: callable = pdel
        if doc is None and fget is not None:
            doc = fget.__doc__
        self.__doc__ = doc
        self.owner = None
        self.name = ''

        self.instance = None

    def __set_name__(self, owner, name):
        self.owner = owner
        self.name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self
        else:
            self.instance = instance

        if self.fget is None:
            raise AttributeError(
                f'IndexableProperty {self.name!r} of {type(instance).__name__!r} object has no getter'
            )
        return self

    def __getitem__(self, item):
        if self.fget is None:
            raise AttributeError(
                f'IndexableProperty {self.name!r} of {type(self.instance).__name__!r} object has no item getter'
            )
        return self.fget(self.instance, item)

    def __setitem__(self, key, value):
        if self.fset is None:
            raise AttributeError(
                f'IndexableProperty {self.name!r} of {type(self.instance).__name__!r} object has no item setter'
            )
        self.fset(self.instance, key, value)

    def __delitem__(self, key):
        if self.fdel is None:
            raise AttributeError(
                f'IndexableProperty {self.name!r} of {type(self.instance).__name__!r} object has no item deleter'
            )
        self.fdel(self.instance, key)

    def __delete__(self, obj):
        if self.pdel is None:
            raise AttributeError(
                f'IndexableProperty {self.name!r} of {type(obj).__name__!r} object has no deleter'
            )
        self.pdel(obj)

    def itemgetter(self, fget: callable):
        """
        Defines the __getitem__ function for this indexable property.
        :param fget: A function that corresponds with the __getitem__ parameter list.
        :return: The indexable property.
        """
        prop = type(self)(fget, self.fset, self.fdel, self.pdel, self.__doc__)

        return prop

    def itemsetter(self, fset: callable):
        """
        Defines the __setitem__ function for this indexable property.
        :param fset: A function that corresponds with the __setitem__ parameter list.
        :return: The indexable property.
        """
        prop = type(self)(self.fget, fset, self.fdel, self.pdel, self.__doc__)

        return prop

    def itemdeleter(self, fdel: callable):
        """
        Defines the __delitem__ function for this indexable property.
        :param fdel: A function that corresponds with the __delitem__ parameter list.
        :return: The indexable property.
        """
        prop = type(self)(self.fget, self.fset, fdel, self.pdel, self.__doc__)

        return prop

    def deleter(self, pdel: callable):
        """
        Defines the __del__ function for this indexable property.
        :param pdel: A function that corresponds with the __del__ parameter list.
        :return: The indexable property.
        """
        prop = type(self)(self.fget, self.fset, self.fdel, pdel, self.__doc__)
        prop.name = self.name
        return prop


class RequiresOptionalImport:
    """
    A decorator class that checks during runtime if an optional import has been imported before running
    the decorated code section.
    """

    def __init__(self, import_: str, from_: str = None, as_: str = None, msg: str = None, check_modules: bool = False):
        """
        Creates a new check for the required optional import during runtime.
        Raises an Error if the specified requirement is not imported when calling the decorated code.

        E.g. if you have an optional import in a try-catch block like:

        import A from b.c as D

        you can decorate a function, class or method as followed:

        @RequiresOptionalImport(import_='A', from_='b.c', as_='D')

        :param import_: The import part of the complete import statement within the document.
        :param from_: The from part of the complete import statement within the document.
        :param as_: The as part of the complete import statement within the document.
        :param msg: The error message to raise.
        :param check_modules: If true also checks if the modules are imported.
        """
        self._check_modules = check_modules

        if self._check_modules:
            module_name = import_ if from_ is None else from_
            all_modules = module_name.split('.')
            self._module_names_to_check = ['.'.join(all_modules[:i + 1]) for i in range(0, len(all_modules))]

        self._reference_name_to_check = import_.split('.')[0] if from_ is None else import_
        self._reference_name_to_check = as_ if as_ is not None else self._reference_name_to_check
        self._msg = f"Unresolved reference '{self._reference_name_to_check}'" if msg is None else msg

    def __call__(self, code):
        """
        :return: Returns a wrapper for a given function.
        """
        if isfunction(code):
            return self._func_wrapper(code)

        elif isclass(code):
            return self._class_wrapper(code)

        raise TypeError(f"argument 'code' must be function or class, not {type(code).__name__}")

    def _func_wrapper(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if self._check_modules:
                for module_name in self._module_names_to_check:
                    if module_name not in sys.modules:
                        raise ModuleNotFoundError(f"No module named '{module_name}'")

            if (self._reference_name_to_check is not None and
                    self._reference_name_to_check not in sys.modules[func.__module__].__dict__):
                raise RuntimeError(self._msg)

            return func(*args, **kwargs)

        return wrapper

    def _class_wrapper(self, cls):
        for key, value in cls.__dict__.items():
            if isfunction(value):
                setattr(cls, key, self._func_wrapper(value))

        return cls


def replace_none_in_slice(item: slice) -> slice:
    """
    Replaces None values in a slice object with default values.
    The default start value is 0, the default stop value is sys.maxsize and the default step value is 1.
    :param item: A slice object.
    :return: A new slice object where None values are replaced by the default values described above.
    """
    new_start = 0 if item.start is None else item.start
    new_stop = sys.maxsize if item.stop is None else item.stop
    new_step = 1 if item.step is None else item.step

    return slice(new_start, new_stop, new_step)


def subtract_offset_from_slice(item: slice, offset: int) -> slice:
    """
    Correctly subtracts an offset from a slice object.
    Note that subtracting a negative offset to a negative start or stop value where the difference exceeds the
    negative number space results in sys.maxsize (start: -1 - offset: -5 => new_start: sys.maxsize).

    E.g. if you have a slice(2, 10, 1) object, and you want to subtract the offset value 5 you get
    a slice(0, 5, 1) object.

    :param item: A slice object.
    :param offset: The offset to subtract.
    :return: A new slice object where the start and stop have been subtracted by offset.
    """
    if offset == 0:
        return item

    if item.start > 0:
        new_start = max(0, item.start - offset)
    else:
        new_start = item.start - offset if item.start - offset < 0 else sys.maxsize
    if item.stop > 0:
        new_stop = max(0, item.stop - offset)
    else:
        new_stop = item.stop - offset if item.stop - offset < 0 else sys.maxsize

    return slice(new_start, new_stop, item.step)


def add_offset_to_slice(item: slice, offset: int) -> slice:
    """
    Correctly adds an offset to a slice object.
    Note that adding a positive offset to a negative start or stop value where the difference exceeds the
    negative number space results in sys.maxsize (start: -1 + offset: 5 => new_start: sys.maxsize).

    E.g. if you have a slice(2, 10, 1) object, and you want to add the offset value 5 you get
    a slice(7, 15, 1) object.

    :param item: A slice object.
    :param offset: The offset to add.
    :return: A new slice object where offset has been added to start and stop.
    """
    return subtract_offset_from_slice(item=item, offset=-offset)



from typing import List, Optional, Iterator

from .raspi_base import Raspi40PinBarData
from ..base import GPIOPinBar, _GPIOPin, GPIOPinState, GPIOPinMode, PinAddressing, PinType
from ..util import RequiresOptionalImport

try:
    import RPi.GPIO as GPIO
    GPIO.setmode(GPIO.BOARD)
except RuntimeError as e:
    print("Error importing RPi.GPIO! This is probably because you need superuser privileges.  "
          "You can achieve this by using 'sudo' to run your script")
    raise
except ModuleNotFoundError as e:
    pass


@RequiresOptionalImport(import_='RPi.GPIO', as_='GPIO')
class RPiGPIOPinBar(GPIOPinBar):
    """
    Pin Bar Wrapper Class for the RPi.GPIO library.
    """

    def __init__(
            self,
            pin_assignment: List[PinType],
            initial_addressing: PinAddressing,
            idx_offset: int = 0,
            gpio_order_from_ids: List[int] = None,
            gpio_idx_offset: int = 0
    ):
        """
        Creates a new RPiGPIOPinBar object which wraps the RPi.GPIO library.
        The object provides functionality for the gpio pins and wraps them to the RPi.GPIO library interface.
        This class supports custom kinds of pin bar layouts but needs the corresponding setup information.
        If you just need a RPi.GPIO pin bar for the 40 pin header of the raspberry simply instantiate an
        Raspi40PinBarRPi object.

        :param pin_assignment: The pin assignment list contains each pin's type.
          The list is ordered by the pin id on the pin bar starting with the lowest id.
        :param initial_addressing: Determines which pins are addressed for given indices.
        :param idx_offset: The id starting offset on the pin bar. This value matches the lowest existing pin bar id.
          E.g. the first pin has the id 1 on the pin bar, then this parameter needs to be 1 too.
        :param gpio_order_from_ids: A list of integers that determines the gpio id enumeration.
          The list is ordered by the gpio id starting with the lowest gpio id and contains the pin bar ids.
          E.g. Pin 27 on the pin bar is labeled as GPIO_0, then the first entry of this list needs to be 27.
        :param gpio_idx_offset: The gpio id starting offset of the individual gpio enumeration.
          This value matches the lowest existing gpio id.
        """
        super().__init__(
            pin_assignment=pin_assignment,
            initial_addressing=initial_addressing,
            idx_offset=idx_offset,
            gpio_order_from_ids=gpio_order_from_ids,
            gpio_idx_offset=gpio_idx_offset,
        )

        # mappings between rpi library and wrapper library
        self._mode_to_rpi_mapping: dict[GPIOPinMode, int] = {
            GPIOPinMode.IN: GPIO.IN,
            GPIOPinMode.IN_PULL_DOWN: GPIO.IN,
            GPIOPinMode.IN_PULL_UP: GPIO.IN,
            GPIOPinMode.OUT: GPIO.OUT,
        }
        self._mode_to_rpi_pull_up_down: dict[GPIOPinMode, Optional[bool]] = {
            GPIOPinMode.IN: GPIO.PUD_OFF,
            GPIOPinMode.IN_PULL_DOWN: GPIO.PUD_DOWN,
            GPIOPinMode.IN_PULL_UP: GPIO.PUD_UP,
            GPIOPinMode.OUT: GPIO.PUD_OFF,
        }
        self._rpi_to_state_mapping: dict[int, GPIOPinState] = {
            GPIO.HIGH: GPIOPinState.HIGH,
            GPIO.LOW: GPIOPinState.LOW,
        }
        self._state_to_rpi_mapping: dict[GPIOPinState, int] = {
            GPIOPinState.HIGH: GPIO.HIGH,
            GPIOPinState.LOW: GPIO.LOW,
        }

    def _change_pin_modes(self, pins: List[_GPIOPin], new_modes_iterator: Iterator[GPIOPinMode]):
        new_modes = [next(new_modes_iterator) for _ in pins]

        pins_and_new_modes = zip(pins, new_modes)
        pins_and_new_modes = [(pin, new_mode) for pin, new_mode in pins_and_new_modes
                              if pin.mode is not new_mode]

        cleanup_pins = [pin for pin, new_mode in pins_and_new_modes if pin.mode is not GPIOPinMode.OFF]
        for pin in cleanup_pins:
            GPIO.cleanup(pin.idx)
            pin.mode = GPIOPinMode.OFF

        setup_pins_and_modes = [(pin, mode) for pin, mode in pins_and_new_modes if mode is not GPIOPinMode.OFF]
        for pin, new_mode in setup_pins_and_modes:
            pin_id = pin.idx
            rpi_mode = self._mode_to_rpi_mapping[new_mode]
            rpi_pull_up_down = self._mode_to_rpi_pull_up_down[new_mode]

            GPIO.setup(pin_id, rpi_mode, pull_up_down=rpi_pull_up_down)
            pin.mode = new_mode

    def _gpio_pin_states_iterator(self, pins: List[_GPIOPin]) -> Iterator[Optional[GPIOPinState]]:
        channel_ids = [pin.idx for pin in pins]
        states: List[int] = [GPIO.input(idx) for idx in channel_ids]
        states: List[GPIOPinState] = [self._rpi_to_state_mapping[state] for state in states]

        return iter(states)

    def _change_gpio_pin_states(self, pins: List[_GPIOPin], new_states: List[GPIOPinState]) -> None:
        channel_ids = [pin.idx for pin in pins]
        states: List[int] = [self._state_to_rpi_mapping[state] for state in new_states]
        GPIO.output(channel_ids, states)

    def __del__(self) -> None:
        self.reset_gpio_pins()


@RequiresOptionalImport(import_='RPi.GPIO', as_='GPIO')
class Raspi40PinBarRPi(RPiGPIOPinBar):
    """
    Implementation of a 40 pin Raspberry pin bar based on the RPi.GPIO library.
    Note that this class needs the RPi.GPIO library to be imported during runtime.
    """

    def __init__(self, initial_addressing: PinAddressing):
        """
        Creates a new instance of a 40 pin Raspberry pin bar using the RPi.GPIO library to communicate with the pi.
        :param initial_addressing: Determines which pins are addressed for given indices.
        """
        super().__init__(
            pin_assignment=Raspi40PinBarData.pin_assignment(),
            initial_addressing=initial_addressing,
            idx_offset=Raspi40PinBarData.idx_offset(),
            gpio_order_from_ids=Raspi40PinBarData.gpio_orders_from_ids(),
            gpio_idx_offset=Raspi40PinBarData.gpio_idx_offset(),
        )



import sys
import unittest
from unittest.mock import patch, MagicMock

from parameterized import parameterized

from gpiowrapper.util import IndexableProperty, RequiresOptionalImport, replace_none_in_slice, \
    subtract_offset_from_slice
from gpiowrapperTests.useful_test_util import ExtendedTestCase


class IndexablePropertyUser:
    def __init__(self, array):
        self._my_list = array

    @IndexableProperty
    def my_property(self, item):
        return self._my_list[item]

    @my_property.itemsetter
    def my_property(self, key, value):
        self._my_list[key] = value


class IndexablePropertyTests(unittest.TestCase):
    def test_get_property__Always__ReturnsIndexablePropertyObject(self):
        # Arrange
        user = IndexablePropertyUser([i for i in range(100)])

        # Act
        actual = user.my_property

        # Assert
        self.assertIsInstance(actual, IndexableProperty)

    def test_set_property__Always__RaisesAttributeError(self):
        # Arrange
        user = IndexablePropertyUser([i for i in range(100)])

        # Act & Act
        with self.assertRaises(AttributeError):
            user.my_property = []

    def test_getitem__All__ReturnsCompleteList(self):
        # Arrange
        user = IndexablePropertyUser([i for i in range(100)])

        # Act
        actual = user.my_property[:]

        # Assert
        self.assertIsInstance(actual, list)
        self.assertListEqual(actual, [i for i in range(100)])

    def test_getitem__Sliced__ReturnsCompleteList(self):
        # Arrange
        user = IndexablePropertyUser([i for i in range(100)])

        # Act
        actual = user.my_property[10:40]

        # Assert
        self.assertIsInstance(actual, list)
        self.assertListEqual(actual, [i for i in range(10, 40)])

    def test_setitem__All__ReturnsCompleteList(self):
        # Arrange
        user = IndexablePropertyUser([i for i in range(100)])

        # Act
        user.my_property[:] = [1 for _ in range(100)]

        # Assert
        self.assertListEqual(user._my_list, [1 for _ in range(100)])

    def test_setitem__Sliced__ReturnsCompleteList(self):
        # Arrange
        user = IndexablePropertyUser([i for i in range(100)])

        # Act
        user.my_property[10:40] = [1 for _ in range(10, 40)]

        # Assert
        self.assertListEqual(user._my_list, [1 if 10 <= i < 40 else i for i in range(100)])


class RequiresOptionalImportTests(ExtendedTestCase):
    def setUp(self):
        if 'a' in sys.modules:
            sys.modules.pop('a')
        if 'b' in sys.modules:
            sys.modules.pop('b')
        if 'c' in sys.modules:
            sys.modules.pop('c')

    def test_ImportStatement__NotImportedAndUnused__NoError(self):
        # Arrange
        assert 'a' not in sys.modules[self.__module__].__dict__

        # Act & Assert
        with self.assertNotRaised():
            @RequiresOptionalImport(import_='a.b.c')
            def method_requires_import():
                pass

    @patch.dict(sys.modules[IndexablePropertyUser.__module__].__dict__, {'a': MagicMock()})
    def test_ImportStatement__ImportedAndUnused__NoError(self):
        # Arrange
        assert 'a' in sys.modules[self.__module__].__dict__

        # Act & Assert
        with self.assertNotRaised():
            @RequiresOptionalImport(import_='a.b.c')
            def method_requires_import():
                pass

    def test_ImportStatement__NotImportedAndUsed__RaisesRuntimeError(self):
        # Arrange
        msg = 'Test MSG'
        assert 'a' not in sys.modules[self.__module__].__dict__

        # Act & Assert
        with self.assertRaisesRegex(RuntimeError, msg):
            @RequiresOptionalImport(import_='a.b.c', msg=msg)
            def method_requires_import():
                pass

            method_requires_import()

    @patch.dict(sys.modules[IndexablePropertyUser.__module__].__dict__, {'a': MagicMock()})
    def test_ImportStatement__ImportedAndUsed__NoError(self):
        # Arrange
        assert 'a' in sys.modules[self.__module__].__dict__

        # Act
        with self.assertNotRaised():
            @RequiresOptionalImport(import_='a.b.c')
            def method_requires_import():
                pass

            method_requires_import()

    @patch.dict(sys.modules[IndexablePropertyUser.__module__].__dict__, {'a': MagicMock()})
    def test_FromImportStatement__NotImportedAndUnused__NoError(self):
        # Arrange
        assert 'd' not in sys.modules[self.__module__].__dict__

        # Act & Assert
        with self.assertNotRaised():
            @RequiresOptionalImport(from_='a.b.c', import_='d')
            def method_requires_import():
                pass

    @patch.dict(sys.modules[IndexablePropertyUser.__module__].__dict__, {'a': MagicMock, 'd': MagicMock()})
    def test_FromImportStatement__ImportedAndUnused__NoError(self):
        # Arrange
        assert 'd' in sys.modules[self.__module__].__dict__

        # Act & Assert
        with self.assertNotRaised():
            @RequiresOptionalImport(from_='a.b.c', import_='d')
            def method_requires_import():
                pass

    @patch.dict(sys.modules[IndexablePropertyUser.__module__].__dict__, {'a': MagicMock()})
    def test_FromImportStatement__NotImportedAndUsed__RaisesRuntimeError(self):
        # Arrange
        msg = 'Test MSG'
        assert 'd' not in sys.modules[self.__module__].__dict__

        # Act & Assert
        with self.assertRaisesRegex(RuntimeError, msg):
            @RequiresOptionalImport(from_='a.b.c', import_='d', msg=msg)
            def method_requires_import():
                pass

            method_requires_import()

    @patch.dict(sys.modules[IndexablePropertyUser.__module__].__dict__, {'a': MagicMock, 'd': MagicMock()})
    def test_FromImportStatement__ImportedAndUsed__NoError(self):
        # Arrange
        assert 'd' in sys.modules[self.__module__].__dict__

        # Act & Assert
        with self.assertNotRaised():
            @RequiresOptionalImport(from_='a.b.c', import_='d')
            def method_requires_import():
                pass

            method_requires_import()

    @patch.dict(sys.modules[IndexablePropertyUser.__module__].__dict__, {'d': MagicMock()})
    def test_FromImportAsStatement__NotImportedAndUnused__NoError(self):
        # Arrange
        assert 'e' not in sys.modules[self.__module__].__dict__

        # Act & Assert
        try:
            @RequiresOptionalImport(from_='a.b.c', import_='d', as_='e')
            def method_requires_import():
                pass
        except Exception as e:
            self.fail(f'Exception thrown! {e}')

    @patch.dict(sys.modules[IndexablePropertyUser.__module__].__dict__, {'d': MagicMock, 'e': MagicMock()})
    def test_FromImportAsStatement__ImportedAndUnused__NoError(self):
        # Arrange
        assert 'e' in sys.modules[self.__module__].__dict__

        # Act & Assert
        with self.assertNotRaised():
            @RequiresOptionalImport(from_='a.b.c', import_='d', as_='e')
            def method_requires_import():
                pass

    @patch.dict(sys.modules[IndexablePropertyUser.__module__].__dict__, {'d': MagicMock()})
    def test_FromImportAsStatement__NotImportedAndUsed__RaisesRuntimeError(self):
        # Arrange
        msg = 'Test MSG'
        assert 'e' not in sys.modules[self.__module__].__dict__

        # Act & Assert
        with self.assertRaisesRegex(RuntimeError, msg):
            @RequiresOptionalImport(from_='a.b.c', import_='d', as_='e', msg=msg)
            def method_requires_import():
                pass

            method_requires_import()

    @patch.dict(sys.modules[IndexablePropertyUser.__module__].__dict__, {'d': MagicMock, 'e': MagicMock()})
    def test_FromImportAsStatement__ImportedAndUsed__NoError(self):
        # Arrange
        assert 'e' in sys.modules[self.__module__].__dict__

        # Act & Assert
        with self.assertNotRaised():
            @RequiresOptionalImport(from_='a.b.c', import_='d', as_='e')
            def method_requires_import():
                pass

            method_requires_import()


class ReplaceNoneInSliceTests(unittest.TestCase):
    def test_replace_none_in_slice__AllNone__SliceOverAllElements(self):
        # Arrange
        item = slice(None, None, None)

        # Act
        new_item = replace_none_in_slice(item=item)

        # Assert
        self.assertEqual(new_item, slice(0, sys.maxsize, 1))

    def test_replace_none_in_slice__StartNone__SetStartToZero(self):
        # Arrange
        item = slice(None, 10, 2)

        # Act
        new_item = replace_none_in_slice(item=item)

        # Assert
        self.assertEqual(new_item, slice(0, 10, 2))

    def test_replace_none_in_slice__StopNone__SetStopToMaxsize(self):
        # Arrange
        item = slice(4, None, 2)

        # Act
        new_item = replace_none_in_slice(item=item)

        # Assert
        self.assertEqual(new_item, slice(4, sys.maxsize, 2))

    def test_replace_none_in_slice__StepNone__SetStepToOne(self):
        # Arrange
        item = slice(4, 30, None)

        # Act
        new_item = replace_none_in_slice(item=item)

        # Assert
        self.assertEqual(new_item, slice(4, 30, 1))


class AddOffsetToSliceTests(unittest.TestCase):
    @parameterized.expand([[slice(3, 6, 1)], [slice(-3, 13, 2)], [slice(13, -3, -1)], [slice(-13, -6, 1)]])
    def test_add_offset_to_slice__WithoutOffset__ReturnsUnchanged(self, item: slice):
        # Arrange
        offset = 0

        # Act
        new_item = subtract_offset_from_slice(item=item, offset=offset)

        # Assert
        self.assertEqual(new_item, item)

    @parameterized.expand([
        [slice(6, 10, 2), slice(1, 5, 2)],
        [slice(4, 10, 1), slice(0, 5, 1)],
        [slice(2, 4, 1), slice(0, 0, 1)],
        [slice(16, 10, -1), slice(11, 5, -1)],
        [slice(-14, -12, 1), slice(-19, -17, 1)],
        [slice(-14, -2, 3), slice(-19, -7, 3)],
        [slice(-4, -2, 1), slice(-9, -7, 1)],
        [slice(-6, -10, -1), slice(-11, -15, -1)],
    ])
    def test_add_offset_to_slice__WithPositiveOffset__ReturnsExpected(self, item, expected):
        # Arrange
        offset = 5

        # Act
        new_item = subtract_offset_from_slice(item=item, offset=offset)

        # Assert
        self.assertEqual(new_item, expected)

    @parameterized.expand([
        [slice(6, 10, 2), slice(11, 15, 2)],
        [slice(4, 10, 1), slice(9, 15, 1)],
        [slice(2, 4, 1), slice(7, 9, 1)],
        [slice(16, 10, -1), slice(21, 15, -1)],
        [slice(-14, -12, 1), slice(-9, -7, 1)],
        [slice(-14, -2, 3), slice(-9, sys.maxsize, 3)],
        [slice(-4, -2, 1), slice(sys.maxsize, sys.maxsize, 1)],
        [slice(-6, -10, -1), slice(-1, -5, -1)],
    ])
    def test_add_offset_to_slice__WithNegativeOffset__ReturnsExpected(self, item, expected):
        # Arrange
        offset = -5

        # Act
        new_item = subtract_offset_from_slice(item=item, offset=offset)

        # Assert
        self.assertEqual(new_item, expected)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, unique
from operator import itemgetter
from types import NoneType
from typing import List, Union, Optional, Any, Iterable, Iterator

from gpiowrapper.util import IndexableProperty, replace_none_in_slice, subtract_offset_from_slice


@unique
class PinAddressing(Enum):
    """
    The pin addressing determines which pin is addressed for a given index.

    - In PinBar mode, the index matches the pin id on the pin bar assignment.
      E.g. if you want to address the pin in the upper left corner you can use the lowest id as
      index assuming that the assignment of the pin ids starts in that corner.
    - In GPIO mode, the index matches the individual gpio pin id enumeration.
      When using this mode you can only address pins of type GPIO.
      E.g. if the 12th pin of the pin bar is labeled as GPIO_0, then you can use the index zero to
      address that pin.
    """
    PinBar = 'PinBar'
    """
    In PinBar mode, the index matches the pin id on the pin bar assignment. E.g. if you want to address the
    pin in the upper left corner you can use the lowest id as index assuming that the assignment of the pin ids 
    starts in that corner.
    """
    GPIO = 'GPIO'
    """
    In GPIO mode, the index matches the individual gpio pin id enumeration. 
    When using this mode you can only address pins of type GPIO. 
    E.g. if the 12th pin of the pin bar is labeled as GPIO_0, then you can use the index zero to 
    address that pin.   
    """

    def __str__(self):
        return self.value


@unique
class PinType(Enum):
    """
    An enumeration of types for pins of a pin bar.
    Entries are:

    - POWER typed pins are responsible for constant power supply.
    - GROUND typed pins can be used to close electronic circuits.
    - GPIO typed pins are general purpose input output pins which are controllable and can be used
      for several kinds of applications.
    - OTHER typed pins with functionality not listed in this enumeration.
    """

    POWER = 0
    """ POWER typed pins are responsible for constant power supply. """
    GROUND = 1
    """ GROUND typed pins can be used to close electronic circuits. """
    GPIO = 2
    """
    GPIO typed pins are general purpose input output pins which are controllable and can be used
    for several kinds of applications.
    """
    OTHER = 3
    """ OTHER typed pins with functionality not listed in this enumeration. """


@unique
class GPIOPinState(Enum):
    """
    Possible gpio pin states.
    Note that this enumeration can only represent discrete values and therefore is incomplete.
    Values are:

    - LOW state represents a voltage-free pin state.
    - HIGH state represents a voltage-high pin state.
    """
    LOW = 0
    """ LOW state represents a voltage-free pin state. """
    HIGH = 1
    """ HIGH state represents a voltage-high pin state. """


@unique
class GPIOPinMode(Enum):
    """
    The Mode of a gpio pin.
    Note that not an implementation may not support all listed modes for a gpio pin.
    Possible modes are:

    - OFF: represents an unused gpio pin.
    - IN: represents an input gpio pin. Note that there is no inbuilt pull down or pull up resistor interposed.
    - IN_PULL_DOWN: represents an input gpio pin with inbuilt pull down resistor. So the default state of a pin in
      this mode is LOW.
    - IN_PULL_UP: represents an input gpio pin with inbuilt pull up resistor. So the default state of a pin in
      this mode is HIGH.
    - OUT: represents an output gpio pin. Note that there is no inbuilt pull down or pull up resistor interposed.
    """
    OFF = 0
    """ represents an unused gpio pin. """
    IN = 1
    """ represents an input gpio pin. Note that there is no inbuilt pull down or pull up resistor interposed. """
    IN_PULL_DOWN = 2
    """
    represents an input gpio pin with inbuilt pull down resistor. So the default state of a pin in this mode is LOW.
    """
    IN_PULL_UP = 3
    """
    represents an input gpio pin with inbuilt pull up resistor. So the default state of a pin in this mode is HIGH.
    """
    OUT = 4
    """ represents an output gpio pin. Note that there is no inbuilt pull down or pull up resistor interposed. """


@dataclass(slots=True, kw_only=True)
class _Pin:
    """
    Dataclass to store pin data.
    """

    idx: int
    """ The id of the pin on the pin bar. """
    type: PinType
    """ The type of the pin. """


@dataclass(slots=True, kw_only=True)
class _GPIOPin(_Pin):
    """
    Dataclass to additionally store gpio pin data.
    """

    gpio_idx: int
    """ The gpio id of the pin. Depends on the individual gpio pin enumeration. """
    mode: GPIOPinMode = GPIOPinMode.OFF
    """ The current mode of the gpio pin. """


class PinBarError(Exception):
    """
    General error type of PinBar.
    All PinBar related errors should be derived from this error class.
    """
    pass


class GPIOPinBarError(PinBarError):
    """
    General error type of GPIOPinBar.
    Superclass of several error types like PinTypeError and ModeIsOffError.
    More precise details must be taken from the individual error message.
    """
    pass


class PinTypeError(PinBarError):
    """
    This error is raised when an attempt is made to apply an operation to a pin that cannot be applied to the pin type.
    E.g. when an attempt is made to change the status of a ground pin.
    """
    pass


class ModeIsOffError(GPIOPinBarError):
    """
    This error is raised when an attempt is made to apply an operation to a GPIO Pin that cannot be applied to the pin
    because the mode of the pin is GPIOPinMode.OFF
    E.g. when attempting to set the state of a gpio pin which is off
    """
    pass


class PinBar(ABC):
    """
    An abstract base class representing an electronic pin bar with zero to several pins.
    """

    def __init__(self, pin_assignment: List[PinType], idx_offset: int = 0):
        """
        Creates a new PinBar object.

        :param pin_assignment: The pin assignment list contains each pin's type.
          The list is ordered by the pin id on the pin bar starting with the lowest id.
        :param idx_offset: The id starting offset on the pin bar. This value matches the lowest existing pin bar id.
          E.g. the first pin has the id 1 on the pin bar, then this parameter needs to be 1 too.
        """
        self._pins: List[_Pin] = [_Pin(idx=i + idx_offset, type=type) for i, type in enumerate(pin_assignment)]
        self._idx_offset: int = idx_offset

    @property
    def pin_assignment(self) -> List[PinType]:
        """
        Gets the pin assignment of this pin bar object as list.

        Note that a change to this list will NOT change the pin assignment of this object.
        To change the pin assignment you need to create a new pin bar object.
        """
        return [pin.type for pin in self._pins]

    @property
    def idx_offset(self) -> int:
        """ Gets the id starting offset of this pin bar object. This value matches the lowest pin id. """
        return self._idx_offset

    def __len__(self) -> int:
        """ Gets the total number of pins for this pin bar. This includes any pin type. """
        return len(self._pins)


class GPIOPinBar(PinBar, ABC):
    """
    An abstract base class representing an electronic pin bar with zero to several pins which provides
    functionality for gpio pins.
    """

    def __init__(
            self,
            pin_assignment: List[PinType],
            initial_addressing: PinAddressing,
            idx_offset: int = 0,
            gpio_order_from_ids: List[int] = None,
            gpio_idx_offset: int = 0
    ):
        """
        Creates a new GPIOPinBar object that provides functionality for gpio pins.

        :param pin_assignment: The pin assignment list contains each pin's type.
          The list is ordered by the pin id on the pin bar starting with the lowest id.
        :param initial_addressing: Determines which pins are addressed for given indices.
        :param idx_offset: The id starting offset on the pin bar. This value matches the lowest existing pin bar id.
          E.g. the first pin has the id 1 on the pin bar, then this parameter needs to be 1 too.
        :param gpio_order_from_ids: A list of integers that determines the gpio id enumeration.
          The list is ordered by the gpio id starting with the lowest gpio id and contains the pin bar ids.
          E.g. Pin 27 on the pin bar is labeled as GPIO_0, then the first entry of this list needs to be 27.
        :param gpio_idx_offset: The gpio id starting offset of the individual gpio enumeration.
          This value matches the lowest existing gpio id.
        """
        super().__init__(pin_assignment=pin_assignment, idx_offset=idx_offset)
        self._gpio_idx_offset: int = gpio_idx_offset

        self._convert_gpio_types_in_pins_list(gpio_order_from_ids)
        self._gpio_pins: List[_GPIOPin] = [pin for pin in self._pins if isinstance(pin, _GPIOPin)]
        self._gpio_pins.sort(key=lambda x: x.gpio_idx)

        # TODO add index and value validator attributes to allow for dynamic validation through subclassing

        self.addressing = initial_addressing

    def _convert_gpio_types_in_pins_list(self, gpio_order_from_ids) -> None:
        """
        Converts pins of type gpio into GPIOPin dataclass objects.
        :param gpio_order_from_ids: see __init__ params.
        """
        gpio_pin_indices = [i for i, pin in enumerate(self._pins) if pin.type is PinType.GPIO]
        gpio_order_indices = gpio_pin_indices

        if gpio_order_from_ids is not None:
            gpio_order_indices = [i - self._idx_offset for i in gpio_order_from_ids]

        if not sorted(gpio_order_indices) == gpio_pin_indices:
            raise ValueError(f'gpio_order_from_ids parameter is invalid! '
                             'Check if you are missing a gpio pin or have duplicates.')

        for i, index in enumerate(gpio_order_indices):
            pin = self._pins[index]
            self._pins[index] = _GPIOPin(
                idx=pin.idx,
                type=PinType.GPIO,
                gpio_idx=i + self._gpio_idx_offset,
                mode=GPIOPinMode.OFF
            )

    @property
    def gpio_idx_offset(self) -> int:
        """
        Gets the gpio id starting offset of this gpio pin bar object.
        This value matches the lowest gpio pin gpio id.
        See PinAddressing.GPIO for more information of the different pin id spaces.
        """
        return self._gpio_idx_offset

    @property
    def num_gpio_pins(self) -> int:
        """ Gets the total number of GPIO Pins. """
        return len(self._gpio_pins)

    @property
    def addressing(self) -> PinAddressing:
        """
        Gets the current pin addressing mode of this pin bar.
        """
        return self._addressing

    @addressing.setter
    def addressing(self, new_addressing: PinAddressing) -> None:
        """
        Sets the pin addressing mode of this pin bar. Allowed addressing modes are PinBar and GPIO.
        :param new_addressing: The new addressing mode.
        """
        if new_addressing not in [PinAddressing.PinBar, PinAddressing.GPIO]:
            raise ValueError(f'{new_addressing} is not supported!')

        self._addressing = new_addressing
        self._addressing_array = self._pins if self.addressing is PinAddressing.PinBar else self._gpio_pins
        self._addressing_offset = self.idx_offset if self.addressing is PinAddressing.PinBar else self.gpio_idx_offset
        self._addressing_array_length = len(self) if self.addressing is PinAddressing.PinBar else self.num_gpio_pins

    @property
    def current_idx_offset(self):
        """
        Returns the idx starting offset for the current PinAddressing.
        :return: The idx_offset for PinAddressing.PinBar or the gpio_idx_offset for PinAddressing.GPIO .
        """
        return self._addressing_offset

    @property
    def current_num_addressable_pins(self):
        """
        Returns the number of pins which are addressable for the current PinAddressing.
        :return: The __len__(self) for PinAddressing.PinBar or the num_gpio_pins for PinAddressing.GPIO .
        """
        return self._addressing_array_length

    class _IndexValidator:
        """ Validator for index items that validates valid types and values. """

        @classmethod
        def _validation_mapping(cls) -> dict:
            return {
                int: cls.validate_int_index,
                slice: cls.validate_slice_index,
                list: cls.validate_list_index,
            }

        @classmethod
        def validate(
                cls,
                item: Any,
                offset: int,
                length: int,
                addressing: PinAddressing  # TODO change to GPIOPinBar object
        ) -> None:
            """
            Validates both type and value of the given item as index.
            :param item: The index item.
            :param offset: The current offset. Depends on the current pin addressing used.
            :param length: The length of the pin array. Depends on the current pin addressing used.
            :param addressing: The current pin addressing used.
            """
            cls.validate_index_type(type(item))
            validation_func = cls._validation_mapping()[type(item)]

            if validation_func is not None:
                # noinspection PyArgumentList
                validation_func(item=item, offset=offset, length=length, addressing=addressing)

        @classmethod
        def validate_index_type(cls, item_type):
            """ Validates the type of the index item. """
            if item_type not in [int, slice, list]:
                raise TypeError(f'index must be integer, slice or list, not {item_type.__name__}')

        @classmethod
        def validate_int_index(cls, item: int, offset: int, length: int, addressing: PinAddressing, **_):
            if not offset <= item < offset + length:
                raise IndexError(f'index out of range for {PinAddressing.__name__}.{str(addressing)}')

        @classmethod
        def validate_slice_index(cls, item: slice, **_):
            if (
                    (item.start is not None and not isinstance(item.start, int)) or
                    (item.stop is not None and not isinstance(item.stop, int)) or
                    (item.step is not None and not isinstance(item.step, int))
            ):
                raise TypeError(f'slice indices must be integers or None or have an __index__ method')

        @classmethod
        def validate_list_index(cls, item: list, offset: int, length: int, addressing: PinAddressing, **_):
            for i, index in enumerate(item):
                if isinstance(index, int):
                    try:
                        cls.validate_int_index(index, offset, length, addressing)
                    except IndexError:
                        raise IndexError(
                            f'index out of range for {PinAddressing.__name__}.{str(addressing)} at position {i}'
                        ) from None
                else:
                    raise TypeError(f'list indices must be integers or have an __index__ method. '
                                    f'Found type {type(index).__name__} at position {i}')

    class _ValueValidator(ABC):
        """
        Abstract base class to validate values set for the gpio pin attributes.
        """
        @classmethod
        def _validation_mapping(cls):
            """ Gets the validation method for each index type. """
            return {
                int: cls._validate_for_int_index,
                slice: cls._validate_for_slice_index,
                list: cls._validate_for_list_index,
            }

        @classmethod
        def validate(cls, value, index_type, considered_pins: List[_Pin]):
            """
            Validates both type and value of the given value.
            :param value: The value to validate.
            :param index_type: The index type used to address the list.
            :param considered_pins: A list of the considered pins of the current operation.
            """
            cls._validate_type(type(value))

            validation_func = cls._validation_mapping()[index_type]
            if validation_func is not None:
                # noinspection PyArgumentList
                validation_func(value=value, considered_pins=considered_pins)

        @classmethod
        @abstractmethod
        def _validate_type(cls, value_type):
            """ Validates the type of the value. """
            pass

        @classmethod
        def _validate_for_int_index(cls, value, **_):
            """ Validates the value for an int index. """
            if isinstance(value, list):
                raise ValueError('setting an array element with a sequence.')

        @classmethod
        def _validate_for_slice_index(cls, value, considered_pins: List[_Pin], **_):
            """ Validates the value for a slice index. """
            if isinstance(value, list):
                cls._validate_list_size(considered_pins, value)
                cls._validate_list_entries(value, considered_pins, none_allowed=True)

        @classmethod
        def _validate_for_list_index(cls, value, considered_pins: List[_Pin], **_):
            """ Validates the value for a list index. """
            if isinstance(value, list):
                cls._validate_list_size(considered_pins, value)
                cls._validate_list_entries(value, considered_pins, none_allowed=False)

        @classmethod
        def _validate_list_size(cls, considered_pins, value):
            """ Validates that the number of values and the number of considered pins are the same. """
            if len(value) is not len(considered_pins):
                raise ValueError(
                    f'could not broadcast input array from size {len(value)} into size {len(considered_pins)}')

        @classmethod
        def _validate_list_entries(cls, value, considered_pins: List[_Pin], none_allowed: bool) -> None:
            """
            Validates that the value list entries are valid for each individual pin type of the
            considered pins.
            :param value: The list of values to validate.
            :param considered_pins: The considered pins by this operation.
            :param none_allowed: If None values are allowed or not.
            """
            value_pin_zip = zip(value, considered_pins)
            allowed_entry_types = cls._allowed_list_entry_types()
            if none_allowed:
                allowed_entry_types.append(NoneType)

            for i, (v, pin) in enumerate(value_pin_zip):
                if type(v) not in allowed_entry_types:
                    allowed_entry_type_names = [i.__name__ for i in allowed_entry_types]
                    msg = (f'list value must be of type {", ".join(allowed_entry_type_names)}. '
                           f'Found type {type(v).__name__} at position {i}')
                    raise ValueError(msg)

                if (v is None and isinstance(pin, _GPIOPin)) or (v is not None and not isinstance(pin, _GPIOPin)):
                    msg = f'unable to assign {v} at position {i} to pin of type {pin.type}'
                    raise PinTypeError(msg)

        @classmethod
        @abstractmethod
        def _allowed_list_entry_types(cls) -> List:
            """ Gets a list of the allowed data types when the value to validate is a list. """
            pass

    class _ModeValueValidator(_ValueValidator):
        """ The value validator when setting new gpio pin mode(s). """
        @classmethod
        def _validate_type(cls, value_type):
            if value_type not in [GPIOPinMode, list]:
                raise TypeError(f'value must be {GPIOPinMode.__name__} or list, not {value_type.__name__}')

        @classmethod
        def _allowed_list_entry_types(cls) -> List:
            return [GPIOPinMode]

    class _StateValueValidator(_ValueValidator):
        """ The value validator when setting new gpio pin state(s). """
        @classmethod
        def _validate_type(cls, value_type):
            if value_type not in [GPIOPinState, list]:
                raise TypeError(f'value must be {GPIOPinState.__name__} or list, not {value_type.__name__}')

        @classmethod
        def _validate_list_entries(cls, value, considered_pins: List[_Pin], none_allowed: bool):
            value_pin_zip = zip(value, considered_pins)
            allowed_entry_types = cls._allowed_list_entry_types()
            if none_allowed:
                allowed_entry_types.append(NoneType)

            for i, (v, pin) in enumerate(value_pin_zip):
                if type(v) not in allowed_entry_types:
                    allowed_entry_type_names = [i.__name__ for i in allowed_entry_types]
                    msg = (f'list value must be of type {", ".join(allowed_entry_type_names)}. '
                           f'Found type {type(v).__name__} at position {i}')
                    raise ValueError(msg)

                '''
                v: GPIOPinState, pin: NoGPIO => PinTypeError
                v: GPIOPinState, pin: GPIO, mode: off => ModeIsOffError
                v: GPIOPinState, pin: GPIO, mode: on => NoError
                v: None, pin: NoGPIO => NoError
                v: None, pin: GPIO, mode: off => NoError
                v: None, pin: GPIO, mode: on => ValueError
                '''

                msg = f'unable to assign {v} at position {i} to pin of type {pin.type}'
                msg = f'{msg} with {pin.mode}' if isinstance(pin, _GPIOPin) else msg

                if v is not None and not isinstance(pin, _GPIOPin):
                    raise PinTypeError(msg)
                elif v is not None and isinstance(pin, _GPIOPin) and pin.mode is GPIOPinMode.OFF:
                    raise ModeIsOffError(msg)
                elif v is None and isinstance(pin, _GPIOPin) and pin.mode is not GPIOPinMode.OFF:
                    raise ValueError(msg)

        @classmethod
        def _allowed_list_entry_types(cls) -> List:
            return [GPIOPinState]

    @IndexableProperty
    def modes(self, item) -> Union[Optional[GPIOPinMode], List[Optional[GPIOPinMode]]]:
        """
        Get the mode(s) of the pins of this pin bar.
        Considers the currently set pin addressing!

        :param item: The index item. Valid indices are of type int, slice or list.
        :return: A GPIOPinMode for int indices,
          a list of GPIOPinMode and None for slice indices where None represents that the corresponding pin is not
          a gpio pin, a list of GPIOPinMode for list indices.
        """
        self._IndexValidator.validate(item=item, offset=self._addressing_offset,
                                      length=self._addressing_array_length, addressing=self.addressing)

        item_pins = self._get_pins(item=item)

        if isinstance(item_pins, list):
            return [pin.mode if isinstance(pin, _GPIOPin) else None for pin in item_pins]
        else:
            return item_pins.mode if isinstance(item_pins, _GPIOPin) else None

    @modes.itemsetter
    def modes(self, key, value: Union[GPIOPinMode, List[Optional[GPIOPinMode]]]) -> None:
        """
        Set new gpio pin mode(s) for the pins of this pin bar.
        Considers the currently set pin addressing!

        :param key: The index item. Valid indices are of type int, slice or list.
        :param value: The new value(s). Note that if you use a slice index and the slice is addressing
          pins that are not of type gpio, you have to pass a None value for these pins.
        """
        self._IndexValidator.validate(item=key, offset=self._addressing_offset,
                                      length=self._addressing_array_length, addressing=self.addressing)

        item_pins = self._get_pins(item=key)
        item_pins = [item_pins] if not isinstance(item_pins, list) else item_pins

        if not isinstance(key, slice):
            self._validate_pins_are_gpio(item_pins)

        self._ModeValueValidator.validate(value=value, index_type=type(key), considered_pins=item_pins)

        if isinstance(value, list):
            value = [v for v in value if v is not None]

        item_pins = [pin for pin in item_pins if isinstance(pin, _GPIOPin)]
        value_it = iter(lambda: value, -1) if isinstance(value, GPIOPinMode) else iter(value)

        self._change_pin_modes(item_pins, value_it)

    def _change_pin_modes(self, pins: List[_GPIOPin], new_modes_iterator: Iterator[GPIOPinMode]):  # TODO iterator? change state uses list
        """
        Change the modes of the given list of pins.

        :param pins: The pins of which the mode is to be changed.
        :param new_modes_iterator: An iterator object which provides the new modes.
        """
        for pin in pins:
            pin.mode = next(new_modes_iterator)

    @IndexableProperty
    def states(self, item) -> Union[Optional[GPIOPinState], List[Optional[GPIOPinState]]]:
        """
        Get the state(s) of the pins of this pin bar.
        Considers the currently set pin addressing!

        :param item: The index item. Valid indices are of type int, slice or list.
        :return: A GPIOPinState for int indices,
          a list of GPIOPinState and None for slice indices where None represents that the corresponding pin is not
          a gpio pin or the gpio pin is in OFF mode, a list of GPIOPinState for list indices.
        """
        self._IndexValidator.validate(item=item, offset=self._addressing_offset,
                                      length=self._addressing_array_length, addressing=self.addressing)

        item_pins = self._get_pins(item=item)
        item_pins = [item_pins] if not isinstance(item_pins, list) else item_pins
        gpio_pins = [pin for pin in item_pins if isinstance(pin, _GPIOPin)]

        if not isinstance(item, slice):
            self._validate_pins_are_gpio(item_pins)
            self._validate_gpio_pin_modes_not_off(gpio_pins)
        else:
            gpio_pins = [pin for pin in gpio_pins if pin.mode is not GPIOPinMode.OFF]

        state_iterator = self._gpio_pin_states_iterator(gpio_pins)
        if isinstance(item, int):
            return next(state_iterator)
        else:
            return [next(state_iterator) if isinstance(pin, _GPIOPin) and pin.mode is not GPIOPinMode.OFF else None
                    for pin in item_pins]

    @abstractmethod
    def _gpio_pin_states_iterator(self, pins: List[_GPIOPin]) -> Iterator[Optional[GPIOPinState]]:
        """
        Gets an iterator which provides the current states of the given pins.
        The state order of the iterator matches the pin order of the given list of pins.

        :param pins: The pins of which the state is to be read.
        :return: An iterator returning the current state values in the correct order.
        """
        pass

    @states.itemsetter
    def states(self, key, value: Union[GPIOPinState, List[Optional[GPIOPinState]]]) -> None:
        """
        Set new gpio pin state(s) for the pins of this pin bar.
        Considers the currently set pin addressing!

        :param key: The index item. Valid indices are of type int, slice or list.
        :param value: The new value(s). Note that if you use a slice index and the slice is addressing
          pins that are either not of type gpio or are in OFF mode, you have to pass a None value for
          these pins.
        """
        self._IndexValidator.validate(item=key, offset=self._addressing_offset,
                                      length=self._addressing_array_length, addressing=self.addressing)

        item_pins = self._get_pins(item=key)
        item_pins = [item_pins] if not isinstance(item_pins, list) else item_pins
        gpio_pins = [pin for pin in item_pins if isinstance(pin, _GPIOPin)]

        if not isinstance(key, slice):
            self._validate_pins_are_gpio(item_pins)
            self._validate_gpio_pin_modes_not_off(gpio_pins)

        self._StateValueValidator.validate(value=value, index_type=type(key), considered_pins=item_pins)

        if isinstance(key, slice):
            gpio_pins = [pin for pin in gpio_pins if pin.mode is not GPIOPinMode.OFF]
        if isinstance(value, GPIOPinState):
            value = [value for _ in gpio_pins]
        elif isinstance(value, list):
            value = [v for v in value if v is not None]

        self._change_gpio_pin_states(gpio_pins, value)

    @abstractmethod
    def _change_gpio_pin_states(self, pins: List[_GPIOPin], new_states: List[GPIOPinState]) -> None:
        """
        Change the states of the given list of pins.

        :param pins: The pins of which the state is to be changed.
        :param new_states: The list with the new pin states.
        """
        pass

    def _get_pins(self, item: Union[int, slice, list]) -> Union[_Pin, List[_Pin]]:
        """
        Gets the pins addressed by a given index item.
        Considers the currently set pin addressing!

        :param item: An index item of type int, slice or list.
        :return: A list of pins.
        """
        if isinstance(item, int):
            return self._addressing_array[item - self._addressing_offset]

        elif isinstance(item, slice):
            item = replace_none_in_slice(item=item)
            item = subtract_offset_from_slice(item=item, offset=self._addressing_offset)
            return self._addressing_array[item]

        elif isinstance(item, list):
            item = [i - self._addressing_offset for i in item]
            pins = itemgetter(*item)(self._addressing_array)
            return list(pins) if isinstance(pins, Iterable) else [pins]

    @staticmethod
    def _validate_pins_are_gpio(item_pins: Iterable[_Pin]) -> None:
        """ Validates that a list of pins contains only pins of type GPIOPin. """
        for pin in item_pins:
            if not isinstance(pin, _GPIOPin):
                raise PinTypeError(f'pin is not of {PinType.__name__}.{PinType.GPIO} for pin idx {pin.idx}')

    @staticmethod
    def _validate_gpio_pin_modes_not_off(gpio_pins: Iterable[_GPIOPin]) -> None:
        """ Validates tha a list of gpio pins contains only pins which are NOT in OFF mode. """
        for pin in gpio_pins:
            if pin.mode is GPIOPinMode.OFF:
                raise ModeIsOffError(f'cannot get state of gpio pin in {GPIOPinMode.OFF} for pin idx {pin.idx}')

    def reset_gpio_pins(self) -> None:
        """ Resets all pin modes of this pin bar. """
        value_it = iter(lambda: GPIOPinMode.OFF, -1)
        self._change_pin_modes(self._gpio_pins, value_it)


class GPIOPinBarEmulator(GPIOPinBar):
    """ An emulator class for a gpio pin bar that provides full control of the pin modes and states. """

    def __init__(
            self,
            pin_assignment: List[PinType],
            initial_addressing: PinAddressing,
            idx_offset: int = 0,
            gpio_order_from_ids: List[int] = None,
            gpio_idx_offset: int = 0
    ):
        """
        Creates a new GPIOPinBarEmulator object that provides functionality for gpio pins and full control
        of pin modes and states.

        :param pin_assignment: The pin assignment list contains each pin's type.
          The list is ordered by the pin id on the pin bar starting with the lowest id.
        :param initial_addressing: Determines which pins are addressed for given indices.
        :param idx_offset: The id starting offset on the pin bar. This value matches the lowest existing pin bar id.
          E.g. the first pin has the id 1 on the pin bar, then this parameter needs to be 1 too.
        :param gpio_order_from_ids: A list of integers that determines the gpio id enumeration.
          The list is ordered by the gpio id starting with the lowest gpio id and contains the pin bar ids.
          E.g. Pin 27 on the pin bar is labeled as GPIO_0, then the first entry of this list needs to be 27.
        :param gpio_idx_offset: The gpio id starting offset of the individual gpio enumeration.
          This value matches the lowest existing gpio id.
        """
        super().__init__(
            pin_assignment=pin_assignment,
            initial_addressing=initial_addressing,
            idx_offset=idx_offset,
            gpio_order_from_ids=gpio_order_from_ids,
            gpio_idx_offset=gpio_idx_offset,
        )
        self._gpio_states: List[Optional[GPIOPinState]] = [None for _ in self._gpio_pins]

    def _change_pin_modes(self, pins: List[_GPIOPin], new_modes_iterator: Iterator[GPIOPinMode]):
        offset = self.gpio_idx_offset
        for pin in pins:
            pin.mode = next(new_modes_iterator)
            current_state = self._gpio_states[pin.gpio_idx - offset]

            if pin.mode is GPIOPinMode.OFF:
                self._gpio_states[pin.gpio_idx - offset] = None
            elif current_state is None:
                self._gpio_states[pin.gpio_idx - offset] = GPIOPinState.LOW

    def _gpio_pin_states_iterator(self, pins: List[_GPIOPin]) -> Iterator[Optional[GPIOPinState]]:
        offset = self.gpio_idx_offset
        return iter([self._gpio_states[pin.gpio_idx - offset] for pin in pins])

    def _change_gpio_pin_states(self, pins: List[_GPIOPin], new_states: List[GPIOPinState]) -> None:
        offset = self.gpio_idx_offset

        for pin, state in zip(pins, new_states):
            self._gpio_states[pin.gpio_idx - offset] = state

    @IndexableProperty
    def emulated_gpio_modes(self, item) -> Union[GPIOPinMode, List[GPIOPinMode]]:
        """
        Get the gpio pin modes of this pin bar in GPIO pin addressing style.
        :param item: The index item. This is no gpio id! Has to match standard list index types.
        :return: A list of gpio pin modes for the addressed gpio pins or a single GPIOPinMode
          value when item is of type int.
        """
        return [pin.mode for pin in self._gpio_pins][item]

    @emulated_gpio_modes.itemsetter
    def emulated_gpio_modes(self, key, value) -> None:
        """
        Set new gpio pin modes without validation. It is your responsibility that this object is in a valid
        state.
        :param key: The index item. This is no gpio id! Has to match standard list index types.
        :param value: The new value(s) to set. Must follow standard list __setitem__ rules.
        """
        relevant_pins = self._gpio_pins[key]
        relevant_pins = relevant_pins if isinstance(relevant_pins, list) else [relevant_pins]
        value_it = iter([value for _ in relevant_pins]) if isinstance(value, GPIOPinMode) else iter(value)

        for pin in relevant_pins:
            pin.mode = next(value_it)

    @IndexableProperty
    def emulated_gpio_states(self, item) -> Union[None, GPIOPinState, List[Optional[GPIOPinState]]]:
        """
        Get the gpio pin states of this pin bar in GPIO pin addressing style.
        :param item: The index item. This is no gpio id! Has to match standard list index types.
        :return: A list of gpio pin states for the addressed gpio pins or a single None or GPIOPinState
          value when item is of type int.
        """
        return self._gpio_states[item]

    @emulated_gpio_states.itemsetter
    def emulated_gpio_states(self, key, value) -> None:
        """
        Set new gpio pin states without validation. It is your responsibility that this object is in a valid
        state.
        :param key: The index item. This is no gpio id! Has to match standard list index types.
        :param value: The new value(s) to set. Must follow standard list __setitem__ rules.
        """
        self._gpio_states[key] = value



import copy
from dataclasses import dataclass

from gpiowrapper.base import PinType, GPIOPinBarEmulator, PinAddressing


@dataclass
class Raspi40PinBarData:
    """
    PinBar constructor data for a 40 pins Raspberry pin bar.
    from https://www.raspberrypi.com/documentation/computers/raspberry-pi.html
    and https://www.raspberrypi.com/documentation/computers/images/GPIO-Pinout-Diagram-2.png
    (Last revisited 05.03.2024)
    """

    @staticmethod
    def pin_assignment():
        value = [
            PinType.POWER, PinType.POWER, PinType.GPIO, PinType.POWER,  # 1  - 4
            PinType.GPIO, PinType.GROUND, PinType.GPIO, PinType.GPIO,  # 5  - 8
            PinType.GROUND, PinType.GPIO, PinType.GPIO, PinType.GPIO,  # 9  - 12
            PinType.GPIO, PinType.GROUND, PinType.GPIO, PinType.GPIO,  # 13 - 16
            PinType.POWER, PinType.GPIO, PinType.GPIO, PinType.GROUND,  # 17 - 20
            PinType.GPIO, PinType.GPIO, PinType.GPIO, PinType.GPIO,  # 21 - 24
            PinType.GROUND, PinType.GPIO, PinType.GPIO, PinType.GPIO,  # 25 - 28
            PinType.GPIO, PinType.GROUND, PinType.GPIO, PinType.GPIO,  # 29 - 32
            PinType.GPIO, PinType.GROUND, PinType.GPIO, PinType.GPIO,  # 33 - 36
            PinType.GPIO, PinType.GPIO, PinType.GROUND, PinType.GPIO,  # 37 - 40
        ]
        return copy.deepcopy(value)

    @staticmethod
    def idx_offset():
        return 1

    @staticmethod
    def gpio_orders_from_ids():
        value = [27, 28, 3, 5, 7, 29, 31, 26, 24, 21, 19, 23, 32,
                 33, 8, 10, 36, 11, 12, 35, 38, 40, 15, 16, 18, 22, 37, 13]
        return copy.deepcopy(value)

    @staticmethod
    def gpio_idx_offset():
        return 0


class Raspi40PinBarEmulator(GPIOPinBarEmulator):
    """
    An emulator for a 40 pin Raspberry pin bar. Supports all functionality of a GPIOPinBarEmulator.
    """

    def __init__(self, initial_addressing: PinAddressing):
        """
        Creates a new instance of an emulator for a 40 pin Raspberry pin bar.
        :param initial_addressing: Determines which pins are addressed for given indices.
        """
        super().__init__(
            pin_assignment=Raspi40PinBarData.pin_assignment(),
            initial_addressing=initial_addressing,
            idx_offset=Raspi40PinBarData.idx_offset(),
            gpio_order_from_ids=Raspi40PinBarData.gpio_orders_from_ids(),
            gpio_idx_offset=Raspi40PinBarData.gpio_idx_offset(),
        )



import re
import sys
from pathlib import Path

from setuptools import setup, find_packages

if sys.version_info < (3, 11):
    print(f'requires at least Python 3.11 to run, got version info:\n {sys.version_info}')
    sys.exit(1)

with open(Path('gpiowrapper', '__init__.py'), encoding='utf-8') as f:
    version = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", f.read(), re.M).group(1)

if version is None:
    raise RuntimeError('Cannot find version information.')

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()


def data_files():
    files = []
    return files


def install_requires():
    requires = []
    return requires


def install_extras_require():
    extra_requires = {
        'raspi': ['RPi.GPIO~=0.7.1'],
        'testing': ['pytest', 'parameterized'],
    }

    all_extra_deps = []
    for value in extra_requires.values():
        all_extra_deps.extend(value)
    extra_requires['all'] = all_extra_deps

    return extra_requires


setup(
    name='gpiowrapper',
    version=version,
    description="A gpio wrapper library",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Rene Ebertowski',
    author_email='r.ebertowski@gmx.de',
    url='https://github.com/reauso/gpiowrapper',
    license='MIT',
    keywords="gpio raspberry raspi",
    python_requires=">=3.11",
    install_requires=install_requires(),
    extras_require=install_extras_require(),
    packages=find_packages(),
    include_package_data=True,
    data_files=data_files(),
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Education',
        'Topic :: Home Automation',
        'Topic :: Other/Nonlisted Topic',
        'Topic :: Software Development :: Libraries',
        'Topic :: System :: Emulators',
        'Topic :: System :: Hardware',
        'Topic :: Utilities',
    ],
)



import unittest
from contextlib import contextmanager


class ExtendedTestCase(unittest.TestCase):
    @contextmanager
    def assertNotRaised(self, exc_type=Exception):
        try:
            yield None
        except exc_type as e:
            raise self.failureException(f'{type(e).__name__} (subclass of {exc_type.__name__}) raised')
